#include <poll.h>      /* For poll()   */
#include <stdio.h>     /* For perror() */
#include <unistd.h>    /* For sleep()  */
#include <errno.h>     /* For errno    */

#include <vector>

using std::vector;

#include "tailtask.hh"
#include "socket.hh"
#include "stdin.hh"

typedef vector <struct TailTask> TailTaskListType;

class Tail
{
    typedef vector <struct Socket>   SocketListType;

    typedef SocketListType::const_iterator SockCit;
    typedef SocketListType::iterator       SockIt;
    
    SocketListType Sockets;
    
    TailTaskListType Tasks;

    StdinType Stdin;
    
    void AddTailTask(const TailTask &task)
    {
        DataPacket tmp;
        tmp.Type          = DataPacket::PacketTypeListen;
        tmp.Peer.Protocol = task.Protocol;
        tmp.Peer.DestPort = task.HeadPort;
        
        Stdin.SendPacket(tmp);
        
        Tasks.push_back(task);
    }
    
    struct pollfd AddPoll(const Socket &sock)
    {
        struct pollfd tmp = { sock.fd, POLLIN, 0 };
        
        if(sock.SendQueue.GetSize() > 0 || sock.Type == Socket::SockTypeConnecting)
        {
            /* We want a chance to speak to the socket */
            tmp.events |= POLLOUT;
        }
        
        return tmp;
    }
    
    void BeaconDied()
    {
        Stdin.Die();
    }
    
    void HandleStdinPacket(const DataPacket &pak)
    {
        /* Handle packets from Beacon */
        
        switch(pak.Type)
        {
            case DataPacket::PacketTypeNothing:
                break;
            
            case DataPacket::PacketTypeOpen:
            {
                /* Params: ConnID, Peer.Protocol, Peer.DestPort */
                
                TailTaskListType::const_iterator i;
                for(i=Tasks.begin(); i!=Tasks.end(); ++i)
                {
                    if(pak.Peer.Protocol == i->Protocol
                    && pak.Peer.DestPort == i->HeadPort)
                    {
                        switch(pak.Peer.Protocol)
                        {
                            case SockProtocolTCP:
                            {
                                Socket sock;
                                sock.BecomeLocalConnecterTCP(i->RealPort);
                                Sockets.push_back(sock);

                                Stdin.AssignConnID(pak.ConnID, sock.Peer);
                                break;
                            }
                            case SockProtocolUDP:
                            {
                                Socket sock;
                                sock.BecomeLocalSenderUDP(i->RealPort);
                                Sockets.push_back(sock);
                                
                                Stdin.AssignConnID(pak.ConnID, sock.Peer);
                                break;
                            }
                        }
                        break;
                    }
                }
                break;
            }
            case DataPacket::PacketTypeData:
            {
                /* Params: Peer, Data */
                
                SockIt i;
                for(i=Sockets.begin(); i!=Sockets.end(); ++i)
                    if(i->Peer == pak.Peer)
                        break;
                
                if(i == Sockets.end())
                {
                    /* This unrecognized connection be. Report it we should? */
                }
                else
                {
                    switch(pak.Peer.Protocol)
                    {
                        case SockProtocolTCP:
                        {
                            i->SendTCP(pak.Data);
                            break;
                        }
                        case SockProtocolUDP:
                        {
                            i->SendUDP(pak.Data);
                            break;
                        }
                    }
                }
                break;
            }
            case DataPacket::PacketTypeClose:
            {
                /* Params: Peer */
                
                SockIt i;
                for(i=Sockets.begin(); i!=Sockets.end(); ++i)
                    if(i->Peer == pak.Peer)
                    {
                        DataPacket pak = i->Die();
                        Stdin.SendPacket(pak);
                        Sockets.erase(i);
                        break;
                    }
                break;
            }
            default:
                /* Weird stuff! */
                break;
        }
    }
    
    void HandlePollStdin(const struct pollfd &poll)
    {
        if(poll.revents & (POLLNVAL | POLLHUP))
        {
            /* Beacon does not want to talk to us anymore! */
            
            BeaconDied();
        }
        if(poll.revents & POLLIN)
        {
            /* Beacon wants to say something! */
            
            DataPacket pak = Stdin.Read(STDIN_FILENO);

            HandleStdinPacket(pak);
        }
    }
    
    void HandlePollStdout(const struct pollfd &poll)
    {
        if(poll.revents & (POLLNVAL | POLLHUP))
        {
            /* Beacon does not want to talk to us anymore! */
            
            BeaconDied();
        }
        if(poll.revents & POLLOUT)
        {
            /* We may now speak to the beacon! */
            
            Stdin.Speak(STDOUT_FILENO);
        }
    }
    
    void HandlePollSocket(const struct pollfd &poll, SockIt i)
    {
        Socket &sock = *i;
        
        if(poll.revents & (POLLNVAL | POLLHUP))
        {
            /* This socket has died! */
            
            if(poll.revents & POLLERR)
                sock.CheckError();
            
            DataPacket pak = sock.Die();
            Stdin.SendPacket(pak);
            
            Sockets.erase(i);
            
            return;
        }
        
        switch(sock.Type)
        {
            case Socket::SockTypeListening:
            {
                /* No such things in Master. */
                break;
            }
            case Socket::SockTypeConnecting:
            {
                if(!(poll.revents & POLLOUT))
                {
                    /* What kind of event is this? */
                    break;
                }
                sock.Type = Socket::SockTypeConnected;
                /* Passthru to SockTypeConnected */
            }
            case Socket::SockTypeConnected:
            {
                switch(sock.Peer.Protocol)
                {
                    case SockProtocolTCP:
                    {
                        if(poll.revents & POLLIN)
                        {
                            DataPacket pak = sock.ReceiveTCP();
                            Stdin.SendPacket(pak);
                        }
                        if(poll.revents & POLLOUT)
                        {
                            sock.SendTCP();
                        }
                        break;
                    }
                    case SockProtocolUDP:
                    {
                        /* UDP sender always sends to one port only. */
                        
                        if(poll.revents & POLLIN)
                        {
                            DataPacket pak = sock.ReceiveUDP();
                            Stdin.SendPacket(pak);
                        }
                        if(poll.revents & POLLOUT)
                        {
                            sock.SendUDP();
                        }
                        break;
                    }
                }
                break;
            }
        } /* switch(sock.type) */
    }
    
    void HandlePoll(const struct pollfd &poll)
    {
        if(poll.fd == STDIN_FILENO)
        {
            HandlePollStdin(poll);
        }
        else if(poll.fd == STDOUT_FILENO)
        {
            HandlePollStdout(poll);
        }
        else
        {
            /* Find which socket this is */
            for(SockIt i=Sockets.begin(); i!=Sockets.end(); ++i)
                if(i->fd == poll.fd)
                {
                    HandlePollSocket(poll, i);
                    break;
                }
        }
    }
    
public:
    Tail(const TailTaskListType &Tasks1)
    {
        TailTaskListType::const_iterator i;
        for(i=Tasks1.begin(); i!=Tasks1.end(); ++i)
            AddTailTask(*i);
        
        while(!Stdin.IsDead())
        {
            /* Add all sockets to the poll struct */
            vector<struct pollfd> polls;

            for(SockCit i=Sockets.begin(); i!=Sockets.end(); ++i)
                polls.push_back(AddPoll(*i));

            struct pollfd stdinfd = { STDIN_FILENO, POLLIN, 0 };
            struct pollfd stdoutfd = { STDOUT_FILENO, 0, 0 };

            if(Stdin.WantSpeak() > 0)
            {
                /* We want a chance to speak to stdout */
                stdoutfd.events |= POLLOUT;
            }
            
            /* Add stdin and stdout to the poll struct */
            
            polls.push_back(stdinfd);
            polls.push_back(stdoutfd);
            
            if(poll(&polls[0], polls.size(), -1) < 0)
            {
                if(errno == EINTR)
                {
                    /* Interrupted system call */
                    continue;
                }
                perror("poll");
                /* FIXME: And now what? */
                sleep(1);
                continue;
            }
            
            for(unsigned a=0; a<polls.size(); ++a)
                HandlePoll(polls[a]);
        }
    }
};

int main(int argc, const char *const *argv)
{
    if(isatty(STDIN_FILENO)
    || isatty(STDOUT_FILENO))
    {
    	
    }

    TailTaskListType TailTasks;
    
    TailTasks.push_back(TailTask(6669, 22,   SockProtocolTCP));
    TailTasks.push_back(TailTask(4665, 4662, SockProtocolUDP));
    
    Tail Run(TailTasks);
}
